==> Ajgu-15.6.25.tar.gz: Ajgu-15.6.25/ajgu/storage/bsddb3.py <==
from pathlib import Path
from itertools import chain

from bsddb3.db import DB
from bsddb3.db import DBEnv
from bsddb3.db import DB_HASH
from bsddb3.db import DB_BTREE
from bsddb3.db import DB_CREATE
from bsddb3.db import DB_INIT_LOG
from bsddb3.db import DB_INIT_TXN
from bsddb3.db import DB_INIT_LOCK
from bsddb3.db import DB_INIT_MPOOL
from bsddb3.db import DB_LOG_AUTO_REMOVE

from ..tools import dumps
from ..tools import loads

from ..elements import Edge
from ..elements import Vertex

from ..managers import EdgeManager
from ..managers import VertexManager

from ..packing import pack
from ..packing import unpack
from ..tools import LazySliceGenerator
from ..exceptions import ElementNotFound


class Transaction:

    def __init__(self, storage, sync):
        if not sync:
            try:  # only supported in bdb 6.1+
                from bsddb3.db import DB_TXN_NOSYNC
            except:
                flags = 0
            else:
                flags = DB_TXN_NOSYNC
        else:
            # noflags ftw!
            flags = 0
        self._txn = storage._env.txn_begin(flags=flags)
        self.vertex = VertexManager(self)
        self.edge = EdgeManager(self)
        self._storage = storage
        self._dirty = set()

    def commit(self):
        for element in self._dirty:
            self._storage._save(element)
        self._txn.commit()

    def rollback(self):
        self._txn.abort()


class BSDDB3:
    """Persistent storage based on Oracle Berkeley Database"""

    def __init__(self, path, cache_size=(1, 0), max_cache_size=(1, 0)):
        self._path = Path(path).absolute()
        # create directory (and its parents) if it doesn't exists
        if not self._path.exists():
            self._path.mkdir(parents=True)

        # init bsddb3
        self._env = DBEnv()
        self._env.set_cache_max(*max_cache_size)
        self._env.set_cachesize(*cache_size)
        flags = (
            DB_CREATE
            | DB_INIT_LOG
            | DB_INIT_TXN
            | DB_INIT_MPOOL
            | DB_INIT_LOCK
        )
        self._env.log_set_config(DB_LOG_AUTO_REMOVE, True)
        self._env.set_lg_max(1024 ** 3)
        self._env.open(
            str(self._path),
            flags,
            0
        )

        # create vertices and edges k/v stores
        def new_store(name, method):
            txn = self._txn()
            flags = DB_CREATE
            elements = DB(self._env)
            elements.open(
                name,
                None,
                method,
                flags,
                0,
                txn=txn._txn
            )
            txn.commit()
            return elements

        self._vertices = new_store('vertices', DB_HASH)
        self._edges = new_store('edges', DB_HASH)
        self._vertices_labels = new_store('vertices_labels', DB_BTREE)
        self._edges_labels = new_store('edges_labels', DB_BTREE)

    def _txn(self, sync=True, **kwargs):
        return Transaction(self, sync)

    def checkpoint(self):
        """Flushes the underlying memory pool,
        writes  a checkpoint record to the log
        and then flushes the log"""
        self._env.txn_checkpoint()

    def close(self):
        self._vertices.close()
        self._edges.close()
        self._env.close()

    def _save(self, element):
        txn = element._manager._txn._txn
        if isinstance(element, Vertex):
            value = (
                element._label,
                element._outgoings,
                element._incomings,
                element.properties.freeze()
            )
            store = self._vertices
            index = pack('SS', element._label, element._identifier)
            self._vertices_labels.put(index.encode('utf-8'), '', txn)
        else:
            value = (
                element._start,
                element._label,
                element._end,
                element.properties.freeze()
            )
            store = self._edges
            index = pack('SS', element._label, element._identifier)
            self._edges_labels.put(index.encode('utf-8'), '', txn)
        key = element._identifier.encode('utf-8')
        value = dumps(value)
        store.put(key, value, txn)

    def lookup(self, txn, database, fmt, *lookup):
        """Look up for entries based on `lookup`.

        The function only support string and integers. This must take into
        account how the queried database is constructed (see below). Querying
        happens over the key and returns the idenfiers.

        Label indexes ``'vertices_labels'`` and ``'edges_labels'`` expects
        only a label name so ``lookup`` equals to ``('some_label',)``.

        Properties indexes ``'vertices_properties'`` and ``'edges_properties'``
        expects an ``attribute_name`` and an ``attribute_value`` which can be
        a ``str`` or an ``int``.

        This can also be used to check for the existence of a entry, for
        instance, it's a quick way given the tuple (name, value, identifier)
        to know if the element associated with ``identifier`` has property
        ``name`` associated with ``value``. It's **probably** quickest than
        lloading the object.

        .. warning:: Don't do funky things with ``*key`` only what is
                     documented is supported (and it's for internal use only)!
        """
        database = getattr(self, database)
        cursor = database.cursor(txn._txn)
        index = pack(fmt[:len(lookup)], *lookup)
        next = cursor.set_range(index.encode('utf-8'))

        while next:
            values, _ = next
            values = unpack(fmt, values)
            # check that it's a good candidate
            if all(map(lambda x: x[0] == x[1], zip(lookup, values))):
                yield values
                next = cursor.next()
            else:
                break

    def create_vertex(self, label, txn):
        """Called by a vertex transaction manager to build a vertex.

        For wrapping a Vertex already in the database one must use
        the constructor."""
        identifier = txn.vertex._random_identifier()
        vertex = Vertex(txn.vertex, identifier, label, dict(), dict(), dict())
        txn.vertex._cache[identifier] = vertex
        vertex._dirty()
        return vertex

    def create_edge(self, start, label, end, txn):
        """Called by a edge transaction manager to build an edge.

        For wrapping a Vertex already in the database one must use
        the constructor."""
        identifier = txn.edge._random_identifier()
        edge = Edge(
            txn.edge,
            identifier,
            start._identifier,
            label,
            end._identifier,
            dict()
        )
        txn.edge._cache[identifier] = edge
        edge._dirty()
        # update start and end
        try:
            start._outgoings[edge._label].append(edge.identifier())
        except KeyError:
            start._outgoings[edge._label] = [edge.identifier()]
        try:
            end._incomings[edge._label].append(edge.identifier())
        except KeyError:
            end._incomings[edge._label] = [edge.identifier()]
        # mark vertices as dirty to be updated
        start._dirty()
        end._dirty()
        return edge

    def _delete(self, element):
        key = element._identifier.encode('utf-8')
        txn = element._manager._txn._txn
        if isinstance(element, Vertex):
            for edge in chain(element.outgoings(), element.incomings()):
                edge.delete()
            # delete this element from the store
            self._vertices.delete(key, txn)
        else:
            start = element.start()
            start._outgoings[element._label].remove(element._identifier)
            start._dirty()
            end = element.end()
            end._incomings[element._label].remove(element._identifier)
            end._dirty()
            # delete this element from the store
            self._edges.delete(key, txn)

    def _element(self, storage, ElementClass, identifier, manager, txn=None):
        try:
            element = manager._cache[identifier]
        except KeyError:
            key = identifier.encode('utf-8')
            element = storage.get(key, txn=manager._txn._txn)
            if not element:
                msg = 'Element with id=%s not found' % identifier
                raise ElementNotFound(msg)
            element = loads(element)
            element = ElementClass(manager, identifier, *element)
            manager._cache[identifier] = element
        return element

    get_vertex = lambda s, *a, **kw: s._element(s._vertices, Vertex, *a, **kw)
    get_edge = lambda s, *a, **kw: s._element(s._edges, Edge, *a, **kw)

    def _slice(self, storage, ElementClass, txn):
        def generator():
            cursor = storage.cursor(txn=txn._txn)
            next = cursor.first()
            while next:
                yield next
                next = cursor.next()

        def wrapper(pair):
            identifier, element = pair
            identifier = identifier.decode('utf-8')
            element = loads(element)
            if ElementClass is Vertex:
                return ElementClass(txn.vertex, identifier, *element)
            else:
                return ElementClass(txn.edge, identifier, *element)

        return LazySliceGenerator(wrapper, generator())
    slice_vertices = lambda s, *a, **kw: s._slice(s._vertices, Vertex, *a, **kw)
    slice_edges = lambda s, *a, **kw: s._slice(s._edges, Edge, *a, **kw)


# Transaction less BSDDB3 storage

class NoTxnTransaction:

    def __init__(self, storage):
        self._txn = None  # pass `None` to bsddb3 methods
        self.vertex = VertexManager(self)
        self.edge = EdgeManager(self)
        self._storage = storage
        self._dirty = set()

    def commit(self):
        for element in self._dirty:
            self._storage._save(element)


class NoTxnBSDDB3(BSDDB3):
    """Transaction less BSDDB3 storage.

    With this storage you use `GraphDatabase` with the regular API.
    The underlying desactivate transaction and ACID support. This allows
    to speedup the database but can lead among other things to broken
    databases in case of error."""

    def __init__(self, path, cache_size=(1, 0), max_cache_size=(1, 0)):
        self._path = Path(path).absolute()
        # create directory (and its parents) if it doesn't exists
        if not self._path.exists():
            self._path.mkdir(parents=True)

        # init bsddb3
        self._env = DBEnv()
        self._env.set_cache_max(*max_cache_size)
        self._env.set_cachesize(*cache_size)
        flags = (
            DB_CREATE
            | DB_INIT_LOG
            | DB_INIT_TXN
            | DB_INIT_MPOOL
            | DB_INIT_LOCK
        )
        self._env.log_set_config(DB_LOG_AUTO_REMOVE, True)
        self._env.set_lg_max(1024 ** 3)
        self._env.open(
            str(self._path),
            flags,
            0
        )

        # create vertices and edges k/v stores
        def new_store(name, method):
            flags = DB_CREATE
            elements = DB(self._env)
            elements.open(
                name,
                None,
                method,
                flags,
                0,
            )
            return elements

        self._vertices = new_store('vertices', DB_HASH)
        self._edges = new_store('edges', DB_HASH)
        self._vertices_labels = new_store('vertices_labels', DB_BTREE)
        self._edges_labels = new_store('edges_labels', DB_BTREE)

    def _txn(self, **kwargs):
        return NoTxnTransaction(self)

==> Ajgu-15.07.25.tar.gz: Ajgu-15.07.25/ajgu/storage/bsddb3.py <==
from pathlib import Path
from itertools import chain

from bsddb3.db import DB
from bsddb3.db import DBEnv
from bsddb3.db import DB_HASH
from bsddb3.db import DB_BTREE
from bsddb3.db import DB_CREATE
from bsddb3.db import DB_INIT_LOG
from bsddb3.db import DB_INIT_TXN
from bsddb3.db import DB_INIT_LOCK
from bsddb3.db import DB_INIT_MPOOL
from bsddb3.db import DB_LOG_AUTO_REMOVE

from ..tools import dumps
from ..tools import loads

from ..elements import Edge
from ..elements import Vertex

from ..managers import EdgeManager
from ..managers import VertexManager

from ..packing import pack
from ..packing import unpack
from ..tools import LazySliceGenerator
from ..exceptions import ElementNotFound


class Transaction:

    def __init__(self, storage, indices, sync):
        if not sync:
            try:  # only supported in bdb 6.1+
                from bsddb3.db import DB_TXN_NOSYNC
            except:
                flags = 0
            else:
                flags = DB_TXN_NOSYNC
        else:
            # noflags ftw!
            flags = 0
        self._txn = storage._env.txn_begin(flags=flags)
        self.vertex = VertexManager(self)
        self.edge = EdgeManager(self)
        self._storage = storage
        self._dirty = set()
        self._indices = indices

    def commit(self):
        for element in self._dirty:
            self._storage._save(element)
        self._txn.commit()

    def rollback(self):
        self._txn.abort()


class BSDDB3:
    """Persistent storage based on Oracle Berkeley Database"""

    def __init__(self, path, cache_size=(1, 0), max_cache_size=(1, 0)):
        self._path = Path(path).absolute()
        # create directory (and its parents) if it doesn't exists
        if not self._path.exists():
            self._path.mkdir(parents=True)

        # init bsddb3
        self._env = DBEnv()
        self._env.set_cache_max(*max_cache_size)
        self._env.set_cachesize(*cache_size)
        flags = (
            DB_CREATE
            | DB_INIT_LOG
            | DB_INIT_TXN
            | DB_INIT_MPOOL
            | DB_INIT_LOCK
        )
        self._env.log_set_config(DB_LOG_AUTO_REMOVE, True)
        self._env.set_lg_max(1024 ** 3)
        self._env.open(
            str(self._path),
            flags,
            0
        )

        # create vertices and edges k/v stores
        def new_store(name, method):
            txn = self._txn(list())
            flags = DB_CREATE
            elements = DB(self._env)
            elements.open(
                name,
                None,
                method,
                flags,
                0,
                txn=txn._txn
            )
            txn.commit()
            return elements

        self._vertices = new_store('vertices', DB_HASH)
        self._edges = new_store('edges', DB_HASH)
        self._vertices_labels = new_store('vertices_labels', DB_BTREE)
        self._edges_labels = new_store('edges_labels', DB_BTREE)
        # property indexes
        self._vertices_index = new_store('vertices_index', DB_BTREE)
        self._edges_index = new_store('edges_index', DB_BTREE)

    def _txn(self, indices, sync=True, *args, **kwargs):
        return Transaction(self, indices, sync)

    def checkpoint(self):
        """Flushes the underlying memory pool,
        writes  a checkpoint record to the log
        and then flushes the log"""
        self._env.txn_checkpoint()

    def close(self):
        self._vertices.close()
        self._edges.close()
        self._env.close()

    def _save(self, element):
        txn = element._manager._txn._txn
        properties = element.properties.freeze()
        identifier = element._identifier
        diff = element.properties.diff()
        if isinstance(element, Vertex):
            # index label
            index = pack('SS', element._label, identifier)
            self._vertices_labels.put(index.encode('utf-8'), '', txn)
            # prepare properties indexing
            element_index = self._vertices_index
            indices = element._manager._txn._indices

            element = (
                element._label,
                element._outgoings,
                element._incomings,
                properties
            )
            store = self._vertices
        else:
            index = pack('SS', element._label, identifier)
            self._edges_labels.put(index.encode('utf-8'), '', txn)
            # prepare properties indexing
            element_index = self._edges_index
            indices = element._manager._txn._indices
            element = (
                element._start,
                element._label,
                element._end,
                properties
            )
            store = self._edges

        # index properties
        keys = set(diff.keys()).intersection(indices)
        for key in keys:
            new, old = diff[key]
            if old and new:
                if isinstance(old, int):
                    # `pack` use the same packing logic for every integer
                    # here we use `q` which is the symbol for signed 64bit
                    # because we need to specify something, but it doesn't
                    # change anything as long as the letter represent some
                    # kind of integer
                    fmt = 'SqS'
                elif isinstance(old, str):
                    fmt = 'SSS'
                else:  # do not index
                    # FIXME: add list support
                    continue
                index = pack(fmt, key, old, identifier)
                element_index.delete(index.encode('utf-8'), txn)

            if new:
                if isinstance(new, int):
                    # same as above
                    fmt = 'SqS'
                elif isinstance(new, str):
                    fmt = 'SSS'
                else:  # do not index
                    # FIXME: add list support
                    continue
                index = pack(fmt, key, new, identifier)
                element_index.put(index.encode('utf-8'), identifier, txn)

        # store element
        key = identifier.encode('utf-8')
        element = dumps(element)
        store.put(key, element, txn)

    def lookup(self, txn, database, fmt, *lookup):
        """Look up for entries based on `lookup`.

        - `database` must be a database string name, one of the following:
               `'_vertices'`, `'_vertices_labels'`, `'_vertices_index'`,
                  `'_edges'`, `'_edges_labels'`, `'_edges_index'`

        - `fmt` must describe the full format of the key most `SS` or `SSS

        - `lookup` only support string and integers. This must take into
        account how the queried database is constructed (see below). Querying
        happens over the database key and returns unpacked matching keys.

        Label indexes ``'vertices_labels'`` and ``'edges_labels'`` expects
        only a label name so ``lookup`` equals to ``('some_label',)``.

        Properties indexes ``'_vertices_index'`` and ``'edges_index'``
        expects an ``attribute_name`` and an ``attribute_value`` which can be
        a ``str`` or an ``int``.

        This can also be used to check for the existence of a entry, for
        instance, it's a quick way given the tuple `(name, value, identifier)`
        to know if the element associated with ``identifier`` has property
        ``name`` associated with ``value``. It's **probably** quickest than
        loading the object.

        .. warning:: Don't do funky things with ``*key`` only what is
                     documented is supported and it's for internal use only!
        """
        database = getattr(self, database)
        cursor = database.cursor(txn._txn)
        index = pack(fmt[:len(lookup)], *lookup)
        next = cursor.set_range(index.encode('utf-8'))

        while next:
            values, _ = next
            values = unpack(fmt, values)
            # check that it's a good candidate
            if all(map(lambda x: x[0] == x[1], zip(lookup, values))):
                yield values
                next = cursor.next()
            else:
                break

    def create_vertex(self, label, txn):
        """Called by a vertex transaction manager to build a vertex.

        For wrapping a Vertex already in the database one must use
        the constructor."""
        identifier = txn.vertex._random_identifier()
        vertex = Vertex(txn.vertex, identifier, label, dict(), dict(), dict())
        txn.vertex._cache[identifier] = vertex
        vertex._dirty()
        return vertex

    def create_edge(self, start, label, end, txn):
        """Called by a edge transaction manager to build an edge.

        For wrapping a Vertex already in the database one must use
        the constructor."""
        identifier = txn.edge._random_identifier()
        edge = Edge(
            txn.edge,
            identifier,
            start._identifier,
            label,
            end._identifier,
            dict()
        )
        txn.edge._cache[identifier] = edge
        edge._dirty()
        # update start and end
        try:
            start._outgoings[edge._label].append(edge.identifier())
        except KeyError:
            start._outgoings[edge._label] = [edge.identifier()]
        try:
            end._incomings[edge._label].append(edge.identifier())
        except KeyError:
            end._incomings[edge._label] = [edge.identifier()]
        # mark vertices as dirty to be updated
        start._dirty()
        end._dirty()
        return edge

    # move this to manager

    def _delete(self, element):
        key = element._identifier.encode('utf-8')
        txn = element._manager._txn._txn
        if isinstance(element, Vertex):
            for edge in chain(element.outgoings(), element.incomings()):
                edge.delete()
            # delete this element from the store
            self._vertices.delete(key, txn)
        else:
            start = element.start()
            start._outgoings[element._label].remove(element._identifier)
            start._dirty()
            end = element.end()
            end._incomings[element._label].remove(element._identifier)
            end._dirty()
            # delete this element from the store
            self._edges.delete(key, txn)

    # FIXME: both _element and _slice can be rewritten with the above `lookup`

    def _element(self, storage, ElementClass, identifier, manager, txn=None):
        try:
            element = manager._cache[identifier]
        except KeyError:
            key = identifier.encode('utf-8')
            element = storage.get(key, txn=manager._txn._txn)
            if not element:
                msg = 'Element with id=%s not found' % identifier
                raise ElementNotFound(msg)
            element = loads(element)
            element = ElementClass(manager, identifier, *element)
            manager._cache[identifier] = element
        return element

    get_vertex = lambda s, *a, **kw: s._element(s._vertices, Vertex, *a, **kw)
    get_edge = lambda s, *a, **kw: s._element(s._edges, Edge, *a, **kw)

    def _slice(self, storage, ElementClass, txn):
        def generator():
            cursor = storage.cursor(txn=txn._txn)
            next = cursor.first()
            while next:
                yield next
                next = cursor.next()

        def wrapper(pair):
            identifier, element = pair
            identifier = identifier.decode('utf-8')
            element = loads(element)
            if ElementClass is Vertex:
                return ElementClass(txn.vertex, identifier, *element)
            else:
                return ElementClass(txn.edge, identifier, *element)

        return LazySliceGenerator(wrapper, generator())
    slice_vertices = lambda s, *a, **kw: s._slice(s._vertices, Vertex, *a, **kw)
    slice_edges = lambda s, *a, **kw: s._slice(s._edges, Edge, *a, **kw)


# Transaction less BSDDB3 storage

class NoTxnTransaction:

    def __init__(self, storage, indices):
        self._txn = None  # pass `None` to bsddb3 methods
        self.vertex = VertexManager(self)
        self.edge = EdgeManager(self)
        self._storage = storage
        self._dirty = set()
        self._indices = indices

    def commit(self):
        for element in self._dirty:
            self._storage._save(element)

    def rollback(self):
        pass


class NoTxnBSDDB3(BSDDB3):
    """Transaction less BSDDB3 storage.

    With this storage you use `GraphDatabase` with the regular API.
    The underlying desactivate transaction and ACID support. This allows
    to speedup the database but can lead among other things to broken
    databases in case of error."""

    def __init__(self, path, cache_size=(1, 0), max_cache_size=(1, 0)):
        self._path = Path(path).absolute()
        # create directory (and its parents) if it doesn't exists
        if not self._path.exists():
            self._path.mkdir(parents=True)

        # init bsddb3
        self._env = DBEnv()
        self._env.set_cache_max(*max_cache_size)
        self._env.set_cachesize(*cache_size)
        flags = (
            DB_CREATE
            | DB_INIT_MPOOL
        )
        self._env.log_set_config(DB_LOG_AUTO_REMOVE, True)
        self._env.set_lg_max(1024 ** 3)
        self._env.open(
            str(self._path),
            flags,
            0
        )

        # create vertices and edges k/v stores
        def new_store(name, method):
            flags = DB_CREATE
            elements = DB(self._env)
            elements.open(
                name,
                None,
                method,
                flags,
                0,
            )
            return elements

        self._vertices = new_store('vertices', DB_HASH)
        self._edges = new_store('edges', DB_HASH)
        self._vertices_labels = new_store('vertices_labels', DB_BTREE)
        self._edges_labels = new_store('edges_labels', DB_BTREE)
        #
        self._vertices_index = new_store('vertices_index', DB_BTREE)
        self._edges_index = new_store('edges_index', DB_BTREE)

    def _txn(self, indices, *args, **kwargs):
        return NoTxnTransaction(self, indices)
